# SPDX-FileCopyrightText: 2023-present Trenton H <rda0128ou@mozmail.com>
#
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from mimetypes import guess_type
from typing import TYPE_CHECKING
from typing import Any
from typing import Generic
from typing import TypeVar
from urllib.parse import quote

from anyio.to_thread import run_sync
from httpx import AsyncClient
from httpx import Client

from tika_client._constants import MIN_COMPRESS_LEN
from tika_client.data_models import TikaResponse

if TYPE_CHECKING:
    from collections.abc import Coroutine
    from pathlib import Path


T = TypeVar("T", bound="Client | AsyncClient")


class BaseResource(ABC, Generic[T]):
    def __init__(self, client: T, *, compress: bool) -> None:
        self.client = client
        self.compress = compress

    def get_content_headers(self, filename: str, disposition: str = "attachment") -> dict[str, str]:
        """
        Given a filename, returns the attachment header.

        Args:
            filename: The filename to encode
            disposition: The disposition of the file, defaults to attachment

        Returns:
            The attachment header
        """
        try:
            # Test if filename is ASCII
            filename.encode("ascii")
        except UnicodeEncodeError:
            # For non-ASCII, provide ASCII fallback and UTF-8 encoded version
            ascii_filename = filename.encode("ascii", "replace").decode("ascii")
            # Replace ? marks from replace encoding with underscore for better readability
            ascii_filename = ascii_filename.replace("?", "_")
            # Escape quotes in ASCII version
            ascii_filename = ascii_filename.replace('"', '\\"')
            # UTF-8 encode the original filename and percent-encode the bytes
            utf8_filename = quote(filename.encode("utf-8"))

            return {
                "Content-Disposition": f'{disposition}; filename="{ascii_filename}"; '
                f"filename*=UTF-8''{utf8_filename}",
            }
        else:
            # If ASCII, we still need to escape quotes
            escaped_filename = filename.replace('"', '\\"')
            return {
                "Content-Disposition": f'{disposition}; filename="{escaped_filename}"',
            }

    @abstractmethod
    def put_multipart(  # pragma: no cover
        self,
        endpoint: str,
        filepath: Path,
        mime_type: str | None = None,
    ) -> Any | Coroutine[Any, Any, Any]:
        """
        Given an endpoint, file and an optional mime type, does a multi-part form
        data upload of the file to the end point.

        Returns the JSON response of the server

        Args:
            endpoint: The endpoint to send the file to
            filepath: The path to the file to send
            mime_type: The mime type of the file to send, if it's not provided, it will be guessed

        Returns:
            The JSON response of the server
        """

    @abstractmethod
    def put_content(  # pragma: no cover
        self,
        endpoint: str,
        content: str | bytes,
        mime_type: str | None = None,
    ) -> Any | Coroutine[Any, Any, Any]:
        """
        Give, an endpoint, content and optional mime type, does an HTTP PUT with the given content.

        Returns the JSON response of the server

        Args:
            endpoint: The endpoint to send the content to
            content: The content to send
            mime_type: The mime type of the content, if it's not provided, it will be guessed

        Returns:
            The JSON response of the server
        """

    @staticmethod
    def decoded_response(resp_json: dict[str, Any]) -> TikaResponse:
        """
        If possible, returns a more detailed class with properties that appear often in this
        mime type.  Otherwise, it's a basically raw data response, but with some helpers
        for processing fields into Python types

        Args:
            resp_json: The JSON response from the server

        Returns:
            The decoded response
        """
        return TikaResponse(resp_json)


class SyncResource(BaseResource[Client]):
    def put_multipart(
        self,
        endpoint: str,
        filepath: Path,
        mime_type: str | None = None,
    ) -> Any:
        """
        Given an endpoint, file and an optional mime type, does a multi-part form
        data upload of the file to the end point.

        Args:
            endpoint: The endpoint to send the file to
            filepath: The path to the file to send
            mime_type: The mime type of the file to send, if it's not provided, it will be guessed

        Returns:
            Returns the JSON response of the server
        """
        with filepath.open("rb") as handler:
            response = self.client.post(
                endpoint,
                files={
                    "upload-file": (
                        filepath.name,
                        handler,
                        mime_type if mime_type else guess_type(filepath.name)[0] or "",
                    ),
                },
                headers=self.get_content_headers(filepath.name),
            )
        response.raise_for_status()
        return response.json()

    def put_content(
        self,
        endpoint: str,
        content: str | bytes,
        mime_type: str | None = None,
    ) -> Any:
        """
        Give, an endpoint, content and optional mime type, does an HTTP PUT with the given content.

        Args:
            endpoint: The endpoint to send the content to
            content: The content to send
            mime_type: The mime type of the content, if it's not provided, it will be guessed

        Returns:
            Returns the JSON response of the server
        """
        content_bytes = content.encode() if isinstance(content, str) else content
        content_length = len(content_bytes)

        headers = {}
        if self.compress and content_length > MIN_COMPRESS_LEN:
            from gzip import compress  # noqa: PLC0415

            content_bytes = compress(content_bytes)
            content_length = len(content_bytes)
            headers["Content-Encoding"] = "gzip"

        headers["Content-Length"] = str(content_length)
        if mime_type is not None:
            headers["Content-Type"] = mime_type

        response = self.client.put(endpoint, content=content_bytes, headers=headers)
        response.raise_for_status()
        return response.json()


class AsyncResource(BaseResource[AsyncClient]):
    async def put_multipart(
        self,
        endpoint: str,
        filepath: Path,
        mime_type: str | None = None,
    ) -> Any:
        """
        Given an endpoint, file and an optional mime type, does a multi-part form
        data upload of the file to the end point.

        Args:
            endpoint: The endpoint to send the file to
            filepath: The path to the file to send
            mime_type: The mime type of the file to send, if it's not provided, it will be guessed

        Returns:
            Returns the JSON response of the server
        """
        with filepath.open("rb") as handler:
            response = await self.client.post(
                endpoint,
                files={
                    "upload-file": (
                        filepath.name,
                        handler,
                        mime_type if mime_type else guess_type(filepath.name)[0] or "",
                    ),
                },
                headers=self.get_content_headers(filepath.name),
            )
        response.raise_for_status()
        return response.json()

    async def put_content(
        self,
        endpoint: str,
        content: str | bytes,
        mime_type: str | None = None,
    ) -> Any:
        """
        Give, an endpoint, content and optional mime type, does an HTTP PUT with the given content.

        Args:
            endpoint: The endpoint to send the content to
            content: The content to send
            mime_type: The mime type of the content, if it's not provided, it will be guessed

        Returns:
            Returns the JSON response of the server
        """
        content_bytes = content.encode() if isinstance(content, str) else content
        content_length = len(content_bytes)

        headers = {}
        if self.compress and content_length > MIN_COMPRESS_LEN:
            from gzip import compress  # noqa: PLC0415

            content_bytes = await run_sync(compress, content_bytes)
            content_length = len(content_bytes)
            headers["Content-Encoding"] = "gzip"

        headers["Content-Length"] = str(content_length)
        if mime_type is not None:
            headers["Content-Type"] = mime_type

        response = await self.client.put(endpoint, content=content_bytes, headers=headers)
        response.raise_for_status()
        return response.json()



# SPDX-FileCopyrightText: 2023-present Trenton H <rda0128ou@mozmail.com>
#
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Final

from tika_client._base import AsyncResource
from tika_client._base import SyncResource

if TYPE_CHECKING:
    from pathlib import Path

    from tika_client.data_models import TikaResponse

ENDPOINT: Final[str] = "/meta"
MULTI_PART_ENDPOINT: Final[str] = f"{ENDPOINT}/form"


class SyncMetadata(SyncResource):
    """
    Handles interaction with the /meta endpoint of a Tika
    server REST API.

    See documentation:
    https://cwiki.apache.org/confluence/display/TIKA/TikaServer#TikaServer-MetadataResource
    """

    def from_file(self, filepath: Path, mime_type: str | None = None) -> TikaResponse:
        """
        PUTs the provided document to the metadata endpoint using multipart
        file encoding.  Optionally can provide the mime type

        Args:
            filepath: The path to the file to be sent to the Tika server
            mime_type: The mime type of the file to be sent to the Tika server

        Returns:
            The JSON response from the Tika server
        """
        resp = self.put_multipart(MULTI_PART_ENDPOINT, filepath, mime_type)
        return self.decoded_response(resp)


class AsyncMetadata(AsyncResource):
    """
    Handles interaction with the /meta endpoint of a Tika
    server REST API.

    See documentation:
    https://cwiki.apache.org/confluence/display/TIKA/TikaServer#TikaServer-MetadataResource
    """

    async def from_file(self, filepath: Path, mime_type: str | None = None) -> TikaResponse:
        """
        PUTs the provided document to the metadata endpoint using multipart
        file encoding.  Optionally can provide the mime type

        Args:
            filepath: The path to the file to be sent to the Tika server
            mime_type: The mime type of the file to be sent to the Tika server

        Returns:
            The JSON response from the Tika server
        """
        resp = await self.put_multipart(MULTI_PART_ENDPOINT, filepath, mime_type)
        return self.decoded_response(resp)



# SPDX-FileCopyrightText: 2023-present Trenton H <rda0128ou@mozmail.com>
#
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations

import logging
from abc import ABC
from abc import abstractmethod
from contextlib import AbstractAsyncContextManager
from contextlib import AbstractContextManager
from functools import cached_property
from typing import TYPE_CHECKING
from typing import Generic
from typing import TypeVar

from httpx import AsyncClient
from httpx import Client

from tika_client.__about__ import __version__
from tika_client._base import AsyncResource
from tika_client._base import SyncResource
from tika_client._resource_meta import AsyncMetadata
from tika_client._resource_meta import SyncMetadata
from tika_client._resource_recursive import AsyncRecursive
from tika_client._resource_recursive import SyncRecursive
from tika_client._resource_tika import AsyncTika
from tika_client._resource_tika import SyncTika

if TYPE_CHECKING:
    from types import TracebackType

T = TypeVar("T", bound="Client | AsyncClient")
R = TypeVar("R", bound="SyncResource | AsyncResource")


class BaseTikaClient(ABC, Generic[T, R]):
    """Base class for Tika clients.

    Args:
        tika_url: The URL of the Tika server
        user_agent: Value to send as the User-Agent header.  Defaults to tika-client/{version}
        timeout: The timeout for the HTTP request
        log_level: The logging level
        compress: Whether to compress the response
    """

    def __init__(
        self,
        tika_url: str,
        user_agent: str = f"tika-client/{__version__}",
        *,
        timeout: float = 30.0,
        log_level: int = logging.ERROR,
        compress: bool = False,
    ):
        self.tika_url = tika_url
        self.timeout = timeout
        self.log_level = log_level
        self.compress = compress
        self.user_agent = user_agent

        logging.getLogger("httpx").setLevel(log_level)
        logging.getLogger("httpcore").setLevel(log_level)

    @cached_property
    def _default_headers(self) -> dict[str, str]:
        headers = {"Accept": "application/json", "User-Agent": self.user_agent}

        if self.compress:
            headers["Accept-Encoding"] = "gzip"

        return headers

    @property
    @abstractmethod
    def client(self) -> T:  # pragma: no cover
        pass

    @property
    @abstractmethod
    def metadata(self) -> R:  # pragma: no cover
        pass

    @property
    @abstractmethod
    def tika(self) -> R:  # pragma: no cover
        pass

    @property
    @abstractmethod
    def rmeta(self) -> R:  # pragma: no cover
        pass


class TikaClient(AbstractContextManager["TikaClient"], BaseTikaClient[Client, SyncResource]):
    def __enter__(self) -> TikaClient:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.client.close()

    @cached_property
    def client(self) -> Client:
        return Client(base_url=self.tika_url, timeout=self.timeout, headers=self._default_headers)

    @cached_property
    def metadata(self) -> SyncMetadata:
        return SyncMetadata(self.client, compress=self.compress)

    @cached_property
    def tika(self) -> SyncTika:
        return SyncTika(self.client, compress=self.compress)

    @cached_property
    def rmeta(self) -> SyncRecursive:
        return SyncRecursive(self.client, compress=self.compress)


class AsyncTikaClient(AbstractAsyncContextManager["AsyncTikaClient"], BaseTikaClient[AsyncClient, AsyncResource]):
    async def __aenter__(self) -> AsyncTikaClient:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.client.aclose()

    @cached_property
    def client(self) -> AsyncClient:
        return AsyncClient(base_url=self.tika_url, timeout=self.timeout, headers=self._default_headers)

    @cached_property
    def metadata(self) -> AsyncMetadata:
        return AsyncMetadata(self.client, compress=self.compress)

    @cached_property
    def tika(self) -> AsyncTika:
        return AsyncTika(self.client, compress=self.compress)

    @cached_property
    def rmeta(self) -> AsyncRecursive:
        return AsyncRecursive(self.client, compress=self.compress)
